@@ -1,5 +1,5 @@
 # Copyright (C) 2013 Ishraq Ibne Ashraf <ishraq@tinkerforge.com>
-# Copyright (C) 2014, 2019 Matthias Bolte <matthias@tinkerforge.com>
+# Copyright (C) 2014, 2019-2020 Matthias Bolte <matthias@tinkerforge.com>
 #
 # Redistribution and use in source and binary forms of this file,
 # with or without modification, are permitted. See the Creative
@@ -29,20 +29,32 @@ use Thread::Queue;
 use Math::BigInt;
 use Tinkerforge::IPConnection;
 use Tinkerforge::Error;
+use Tinkerforge::DeviceDisplayNames qw(get_device_display_name);
 
 # constants
 use constant _RESPONSE_EXPECTED_ALWAYS_TRUE => 1; # getter
 use constant _RESPONSE_EXPECTED_TRUE => 2; # setter
 use constant _RESPONSE_EXPECTED_FALSE => 3; # setter,default
 
+use constant _DEVICE_IDENTIFIER_CHECK_PENDING => 0;
+use constant _DEVICE_IDENTIFIER_CHECK_MATCH => 1;
+use constant _DEVICE_IDENTIFIER_CHECK_MISMATCH => 2;
+
 # the constructor
 sub _new
 {
-	my ($class, $uid, $ipcon, $api_version) =  @_;
+	my ($class, $uid, $ipcon, $api_version, $device_identifier, $device_display_name) = @_;
 
-	my $self :shared = shared_clone({uid => _base58_decode($uid),
+	my $self :shared = shared_clone({replaced => shared_clone(0),
+	                                 uid => _base58_decode($uid),
+	                                 uid_string => shared_clone($uid),
 	                                 ipcon => shared_clone($ipcon),
 	                                 api_version => shared_clone($api_version),
+	                                 device_identifier => shared_clone($device_identifier),
+	                                 device_display_name => shared_clone($device_display_name),
+	                                 device_identifier_lock_ref => undef,
+	                                 device_identifier_check => shared_clone(&_DEVICE_IDENTIFIER_CHECK_PENDING),
+	                                 wrong_device_display_name => shared_clone('?'),
 	                                 registered_callbacks => shared_clone({}),
 	                                 callback_formats => shared_clone({}),
 	                                 high_level_callbacks => shared_clone({}),
@@ -51,20 +63,21 @@ sub _new
 	                                 response_queue => Thread::Queue->new(),
 	                                 device_lock_ref => undef,
 	                                 request_lock_ref => undef,
-	                                 response_expected => shared_clone({})});
+	                                 response_expected => shared_clone({}),
+	                                 stream_lock_ref => undef});
 
 	bless($self, $class);
 
+	my $device_identifier_lock :shared;
 	my $device_lock :shared;
 	my $request_lock :shared;
 	my $stream_lock :shared;
 
+	$self->{device_identifier_lock_ref} = \$device_identifier_lock;
 	$self->{device_lock_ref} = \$device_lock;
 	$self->{request_lock_ref} = \$request_lock;
 	$self->{stream_lock_ref} = \$stream_lock;
 
-	$self->{ipcon}->{devices}->{$self->{uid}} = $self;
-
 	return $self;
 }
 
@@ -118,7 +131,7 @@ sub _base58_decode
 
 sub _send_request
 {
-	my ($self, $function_id, $data, $form_data, $form_return) = @_;
+	my ($self, $function_id, $data, $form_data, $length_return, $form_return) = @_;
 
 	lock(${$self->{request_lock_ref}});
 
@@ -215,7 +228,7 @@ sub _send_request
 
 	if($length > 8) # means there is data in payload
 	{
-		$packet = $packet_header.$packed_data;
+		$packet = $packet_header . $packed_data;
 	}
 	else # means no data in payload
 	{
@@ -234,30 +247,37 @@ sub _send_request
 		{
 			if(defined($form_return))
 			{
-				if(length($response_packet) >= 8)
+				if(length($response_packet) >= 8) # FIXME
 				{
 					my $_err_code = $self->{ipcon}->_get_err_from_data($response_packet);
 
-					if($_err_code != 0)
+					if($_err_code == 0)
 					{
-						my $_fid = $self->{ipcon}->_get_fid_from_data($response_packet);
+						if ($length_return == 0)
+						{
+							# Setter with response-expected enabled
+							$length_return = 8;
+						}
 
-						if($_err_code == 1)
+						my $response_packet_length = length($response_packet);
+
+						if ($response_packet_length != $length_return)
 						{
-							croak(Tinkerforge::Error->_new(Tinkerforge::Error->INVALID_PARAMETER, "Got invalid parameter for function $_fid"));
-							return 1;
-						}
-						elsif($_err_code == 2)
-						{
-							croak(Tinkerforge::Error->_new(Tinkerforge::Error->FUNCTION_NOT_SUPPORTED, "Function $_fid is not supported"));
-							return 1;
-						}
-						else
-						{
-							croak(Tinkerforge::Error->_new(Tinkerforge::Error->UNKNOWN_ERROR, "Function $_fid returned an unknown error"));
-							return 1;
+							croak(Tinkerforge::Error->_new(Tinkerforge::Error->WRONG_RESPONSE_LENGTH, "Expected response of $length_return byte for function ID $function_id, got $response_packet_length byte instead"));
 						}
 					}
+					elsif($_err_code == 1)
+					{
+						croak(Tinkerforge::Error->_new(Tinkerforge::Error->INVALID_PARAMETER, "Got invalid parameter for function $function_id"));
+					}
+					elsif($_err_code == 2)
+					{
+						croak(Tinkerforge::Error->_new(Tinkerforge::Error->FUNCTION_NOT_SUPPORTED, "Function $function_id is not supported"));
+					}
+					else
+					{
+						croak(Tinkerforge::Error->_new(Tinkerforge::Error->UNKNOWN_ERROR, "Function $function_id returned an unknown error"));
+					}
 
 					my $response_packet_payload = $self->{ipcon}->_get_payload_from_data($response_packet);
 
@@ -274,6 +294,47 @@ sub _send_request
 	return 1;
 }
 
+sub _check_validity
+{
+	my ($self) = @_;
+
+	if($self->{replaced})
+	{
+		croak(Tinkerforge::Error->_new(Tinkerforge::Error->DEVICE_REPLACED, "Device has been replaced"));
+	}
+
+	if($self->{device_identifier_check} == &_DEVICE_IDENTIFIER_CHECK_MATCH)
+	{
+		return 1;
+	}
+
+	lock(${$self->{device_identifier_lock_ref}});
+
+	if($self->{device_identifier_check} == &_DEVICE_IDENTIFIER_CHECK_PENDING)
+	{
+		my ($uid, $connected_uid, $position, $hardware_version, $firmware_version, $device_identifier, $enumeration_type) =
+		    $self->_send_request(255, [], '', 33, 'Z8 Z8 a C3 C3 S'); # get_identity
+
+		if($device_identifier == $self->{device_identifier})
+		{
+			$self->{device_identifier_check} = shared_clone(&_DEVICE_IDENTIFIER_CHECK_MATCH);
+		}
+		else
+		{
+			$self->{device_identifier_check} = shared_clone(&_DEVICE_IDENTIFIER_CHECK_MISMATCH);
+			$self->{wrong_device_display_name} = get_device_display_name($device_identifier);
+		}
+	}
+
+	if($self->{device_identifier_check} == &_DEVICE_IDENTIFIER_CHECK_MISMATCH)
+	{
+		croak(Tinkerforge::Error->_new(Tinkerforge::Error->WRONG_DEVICE_TYPE,
+		      "UID $self->{uid_string} belongs to a $self->{wrong_device_display_name} instead of the expected $self->{device_display_name}"));
+	}
+
+	return 1;
+}
+
 =head1 FUNCTIONS
 
 =over
@@ -322,7 +383,7 @@ it is disabled by default and can be enabled.
 Enabling the response expected flag for a setter function allows to
 detect timeouts and other error conditions calls of this setter as
 well. The device will then send a response for this purpose. If this
-flag is disabled for a setter function then no response is send and
+flag is disabled for a setter function then no response is sent and
 errors are silently ignored, because they cannot be detected.
 
 =cut
@@ -359,7 +420,7 @@ Changes the response expected flag of the function specified by the
 Enabling the response expected flag for a setter function allows to
 detect timeouts and other error conditions calls of this setter as
 well. The device will then send a response for this purpose. If this
-flag is disabled for a setter function then no response is send and
+flag is disabled for a setter function then no response is sent and
 errors are silently ignored, because they cannot be detected.
 
 =cut
